@@ -5,16 +5,20 @@ use asynchronous_codec as futures_codec;
 use futures::io::{AsyncRead, AsyncReadExt, AsyncWrite, AsyncWriteExt};
 use futures::sink::SinkExt;
 use futures::stream::StreamExt;
+use tor_error::internal;
 
-use crate::channel::codec::ChannelCodec;
+use crate::channel::codec::{ChannelCodec, CodecError};
 use crate::channel::UniqId;
+use crate::util::skew::ClockSkew;
 use crate::{Error, Result};
 use tor_cell::chancell::{msg, ChanCmd};
 
 use std::net::SocketAddr;
+use std::sync::Arc;
+use std::time::SystemTime;
 
 use tor_bytes::Reader;
-use tor_linkspec::ChanTarget;
+use tor_linkspec::{ChanTarget, OwnedChanTarget};
 use tor_llcrypto as ll;
 use tor_llcrypto::pk::ed25519::Ed25519Identity;
 use tor_llcrypto::pk::rsa::RsaIdentity;
@@ -59,6 +63,8 @@ pub struct UnverifiedChannel<T: AsyncRead + AsyncWrite + Send + Unpin + 'static>
     /// The netinfo cell that we got from the relay.
     #[allow(dead_code)] // Relays will need this.
     netinfo_cell: msg::Netinfo,
+    /// How much clock skew did we detect in this handshake?
+    clock_skew: ClockSkew,
     /// Logging identifier for this stream.  (Used for logging only.)
     unique_id: UniqId,
 }
@@ -85,6 +91,15 @@ pub struct VerifiedChannel<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> {
     rsa_id: RsaIdentity,
 }
 
+/// Convert a CodecError to an Error, under the context that it occurs while
+/// doing a channel handshake.
+fn codec_err_to_handshake(err: CodecError) -> Error {
+    match err {
+        CodecError::Io(e) => Error::HandshakeIoErr(Arc::new(e)),
+        CodecError::Cell(e) => Error::HandshakeProto(format!("Invalid cell on handshake: {}", e)),
+    }
+}
+
 impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake<T> {
     /// Construct a new OutboundClientHandshake.
     pub(crate) fn new(tls: T, target_addr: Option<SocketAddr>) -> Self {
@@ -97,7 +112,18 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
 
     /// Negotiate a link protocol version with the relay, and read
     /// the relay's handshake information.
-    pub async fn connect(mut self) -> Result<UnverifiedChannel<T>> {
+    ///
+    /// Takes a function that reports the current time.  In theory, this can just be
+    /// `SystemTime::now()`.
+    pub async fn connect<F>(mut self, now_fn: F) -> Result<UnverifiedChannel<T>>
+    where
+        F: FnOnce() -> SystemTime,
+    {
+        /// Helper: wrap an IoError as a HandshakeIoErr.
+        fn io_err_to_handshake(err: std::io::Error) -> Error {
+            Error::HandshakeIoErr(Arc::new(err))
+        }
+
         match self.target_addr {
             Some(addr) => debug!("{}: starting Tor handshake with {}", self.unique_id, addr),
             None => debug!("{}: starting Tor handshake", self.unique_id),
@@ -108,26 +134,37 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
             let my_versions = msg::Versions::new(LINK_PROTOCOLS)?;
             self.tls
                 .write_all(&my_versions.encode_for_handshake())
-                .await?;
-            self.tls.flush().await?;
+                .await
+                .map_err(io_err_to_handshake)?;
+            self.tls.flush().await.map_err(io_err_to_handshake)?;
         }
+        let versions_flushed_at = coarsetime::Instant::now();
+        let versions_flushed_wallclock = now_fn();
 
         // Get versions cell.
         trace!("{}: waiting for versions", self.unique_id);
         let their_versions: msg::Versions = {
             // TODO: this could be turned into another function, I suppose.
             let mut hdr = [0_u8; 5];
-            let not_relay = || Err(Error::ChanProto("Doesn't seem to be a tor relay".into()));
+            let not_relay = || {
+                Err(Error::HandshakeProto(
+                    "Doesn't seem to be a tor relay".into(),
+                ))
+            };
             match self.tls.read_exact(&mut hdr).await {
                 Err(e) if e.kind() == std::io::ErrorKind::UnexpectedEof => return not_relay(),
                 otherwise => otherwise,
-            }?;
+            }
+            .map_err(io_err_to_handshake)?;
             if hdr[0..3] != [0, 0, ChanCmd::VERSIONS.into()] {
                 return not_relay();
             }
             let msglen = u16::from_be_bytes(*array_ref![hdr, 3, 2]);
             let mut msg = vec![0; msglen as usize];
-            self.tls.read_exact(&mut msg).await?;
+            self.tls
+                .read_exact(&mut msg)
+                .await
+                .map_err(io_err_to_handshake)?;
             let mut reader = Reader::from_slice(&msg);
             reader.extract()?
         };
@@ -136,7 +173,7 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
         // Determine which link protocol we negotiated.
         let link_protocol = their_versions
             .best_shared_link_protocol(LINK_PROTOCOLS)
-            .ok_or_else(|| Error::ChanProto("No shared link protocols".into()))?;
+            .ok_or_else(|| Error::HandshakeProto("No shared link protocols".into()))?;
         trace!("{}: negotiated version {}", self.unique_id, link_protocol);
 
         // Now we can switch to using a "Framed". We can ignore the
@@ -147,14 +184,14 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
 
         // Read until we have the netinfo cells.
         let mut certs: Option<msg::Certs> = None;
-        let mut netinfo: Option<msg::Netinfo> = None;
+        let mut netinfo: Option<(msg::Netinfo, coarsetime::Instant)> = None;
         let mut seen_authchallenge = false;
 
         // Loop: reject duplicate and unexpected cells
         trace!("{}: waiting for rest of handshake.", self.unique_id);
         while let Some(m) = tls.next().await {
             use msg::ChanMsg::*;
-            let (_, m) = m?.into_circid_and_msg();
+            let (_, m) = m.map_err(codec_err_to_handshake)?.into_circid_and_msg();
             trace!("{}: received a {} cell.", self.unique_id, m.cmd());
             match m {
                 // Are these technically allowed?
@@ -164,13 +201,13 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
                 // Clients don't care about AuthChallenge
                 AuthChallenge(_) => {
                     if seen_authchallenge {
-                        return Err(Error::ChanProto("Duplicate authchallenge cell".into()));
+                        return Err(Error::HandshakeProto("Duplicate authchallenge cell".into()));
                     }
                     seen_authchallenge = true;
                 }
                 Certs(c) => {
                     if certs.is_some() {
-                        return Err(Error::ChanProto("Duplicate certs cell".into()));
+                        return Err(Error::HandshakeProto("Duplicate certs cell".into()));
                     }
                     certs = Some(c);
                 }
@@ -178,16 +215,16 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
                     if netinfo.is_some() {
                         // This should be impossible, since we would
                         // exit this loop on the first netinfo cell.
-                        return Err(Error::InternalError(
-                            "Somehow tried to record a duplicate NETINFO cell".into(),
-                        ));
+                        return Err(Error::from(internal!(
+                            "Somehow tried to record a duplicate NETINFO cell"
+                        )));
                     }
-                    netinfo = Some(n);
+                    netinfo = Some((n, coarsetime::Instant::now()));
                     break;
                 }
                 // No other cell types are allowed.
                 m => {
-                    return Err(Error::ChanProto(format!(
+                    return Err(Error::HandshakeProto(format!(
                         "Unexpected cell type {}",
                         m.cmd()
                     )))
@@ -197,15 +234,28 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
 
         // If we have certs and netinfo, we can finish authenticating.
         match (certs, netinfo) {
-            (Some(_), None) => Err(Error::ChanProto("Missing netinfo or closed stream".into())),
-            (None, _) => Err(Error::ChanProto("Missing certs cell".into())),
-            (Some(certs_cell), Some(netinfo_cell)) => {
+            (Some(_), None) => Err(Error::HandshakeProto(
+                "Missing netinfo or closed stream".into(),
+            )),
+            (None, _) => Err(Error::HandshakeProto("Missing certs cell".into())),
+            (Some(certs_cell), Some((netinfo_cell, netinfo_rcvd_at))) => {
                 trace!("{}: received handshake, ready to verify.", self.unique_id);
+                let clock_skew = if let Some(netinfo_timestamp) = netinfo_cell.timestamp() {
+                    let delay = netinfo_rcvd_at - versions_flushed_at;
+                    ClockSkew::from_handshake_timestamps(
+                        versions_flushed_wallclock,
+                        netinfo_timestamp,
+                        delay.into(),
+                    )
+                } else {
+                    ClockSkew::None
+                };
                 Ok(UnverifiedChannel {
                     link_protocol,
                     tls,
                     certs_cell,
                     netinfo_cell,
+                    clock_skew,
                     target_addr: self.target_addr,
                     unique_id: self.unique_id,
                 })
@@ -215,6 +265,14 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> OutboundClientHandshake
 }
 
 impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
+    /// Return the reported clock skew from this handshake.
+    ///
+    /// Note that the skew reported by this function might not be "true": the
+    /// relay might have its clock set wrong, or it might be lying to us.
+    pub fn clock_skew(&self) -> ClockSkew {
+        self.clock_skew
+    }
+
     /// Validate the certificates and keys in the relay's handshake.
     ///
     /// 'peer' is the peer that we want to make sure we're connecting to.
@@ -244,10 +302,39 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
         self,
         peer: &U,
         peer_cert_sha256: &[u8],
-        now: Option<std::time::SystemTime>,
+        now: Option<SystemTime>,
     ) -> Result<VerifiedChannel<T>> {
         use tor_cert::CertType;
         use tor_checkable::*;
+
+        /// Helper: given a time-bound input, give a result reflecting its
+        /// validity at `now`, and the inner object.
+        ///
+        /// We use this here because we want to validate the whole handshake
+        /// regardless of whether the certs are expired, so we can determine
+        /// whether we got a plausible handshake with a skewed partner, or
+        /// whether the handshake is definitely bad.
+        fn check_timeliness<C, T>(checkable: C, now: SystemTime, skew: ClockSkew) -> (Result<()>, T)
+        where
+            C: Timebound<T, Error = TimeValidityError>,
+        {
+            let status = checkable.is_valid_at(&now).map_err(|e| match (e, skew) {
+                (TimeValidityError::Expired(expired_by), ClockSkew::Fast(skew))
+                    if expired_by < skew =>
+                {
+                    Error::HandshakeCertsExpired { expired_by, skew }
+                }
+                // As it so happens, we don't need to check for this case, since the certs in use
+                // here only have an expiration time in them.
+                // (TimeValidityError::NotYetValid(_), ClockSkew::Slow(_)) => todo!(),
+                (_, _) => Error::HandshakeProto("Certificate expired or not yet valid".into()),
+            });
+            let cert = checkable.dangerously_assume_timely();
+            (status, cert)
+        }
+        // Replace 'now' with the real time to use.
+        let now = now.unwrap_or_else(SystemTime::now);
+
         // We need to check the following lines of authentication:
         //
         // First, to bind the ed identity to the channel.
@@ -262,8 +349,20 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
         //    peer.ed_identity().
 
         let c = &self.certs_cell;
-        let id_sk = c.parse_ed_cert(CertType::IDENTITY_V_SIGNING)?;
-        let sk_tls = c.parse_ed_cert(CertType::SIGNING_V_TLS_CERT)?;
+        /// Helper: get a cert from a Certs cell, and convert errors appropriately.
+        fn get_cert(
+            certs: &tor_cell::chancell::msg::Certs,
+            tp: CertType,
+        ) -> Result<tor_cert::KeyUnknownCert> {
+            match certs.parse_ed_cert(tp) {
+                Ok(c) => Ok(c),
+                Err(tor_cell::Error::ChanProto(e)) => Err(Error::HandshakeProto(e)),
+                Err(e) => Err(Error::HandshakeProto(e.to_string())),
+            }
+        }
+
+        let id_sk = get_cert(c, CertType::IDENTITY_V_SIGNING)?;
+        let sk_tls = get_cert(c, CertType::SIGNING_V_TLS_CERT)?;
 
         let mut sigs = Vec::new();
 
@@ -272,20 +371,17 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
         // Check the identity->signing cert
         let (id_sk, id_sk_sig) = id_sk.check_key(&None)?.dangerously_split()?;
         sigs.push(&id_sk_sig);
-        let id_sk = id_sk
-            .check_valid_at_opt(now)
-            .map_err(|_| Error::ChanProto("Certificate expired or not yet valid".into()))?;
+        let (id_sk_timeliness, id_sk) = check_timeliness(id_sk, now, self.clock_skew);
 
         // Take the identity key from the identity->signing cert
         let identity_key = id_sk.signing_key().ok_or_else(|| {
-            Error::ChanProto("Missing identity key in identity->signing cert".into())
+            Error::HandshakeProto("Missing identity key in identity->signing cert".into())
         })?;
 
         // Take the signing key from the identity->signing cert
-        let signing_key = id_sk
-            .subject_key()
-            .as_ed25519()
-            .ok_or_else(|| Error::ChanProto("Bad key type in identity->signing cert".into()))?;
+        let signing_key = id_sk.subject_key().as_ed25519().ok_or_else(|| {
+            Error::HandshakeProto("Bad key type in identity->signing cert".into())
+        })?;
 
         // Now look at the signing->TLS cert and check it against the
         // peer certificate.
@@ -293,12 +389,10 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
             .check_key(&Some(*signing_key))? // TODO(nickm): this is a bad interface
             .dangerously_split()?;
         sigs.push(&sk_tls_sig);
-        let sk_tls = sk_tls
-            .check_valid_at_opt(now)
-            .map_err(|_| Error::ChanProto("Certificate expired or not yet valid".into()))?;
+        let (sk_tls_timeliness, sk_tls) = check_timeliness(sk_tls, now, self.clock_skew);
 
         if peer_cert_sha256 != sk_tls.subject_key().as_bytes() {
-            return Err(Error::ChanProto(
+            return Err(Error::HandshakeProto(
                 "Peer cert did not authenticate TLS cert".into(),
             ));
         }
@@ -309,7 +403,7 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
         // and call pk::validate_all_sigs() instead, but that doesn't gain
         // any performance.
         if !ll::pk::ed25519::validate_batch(&sigs[..]) {
-            return Err(Error::ChanProto(
+            return Err(Error::HandshakeProto(
                 "Invalid ed25519 signature in handshake".into(),
             ));
         }
@@ -326,7 +420,7 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
         let pkrsa = c
             .cert_body(CertType::RSA_ID_X509)
             .and_then(ll::util::x509_extract_rsa_subject_kludge)
-            .ok_or_else(|| Error::ChanProto("Couldn't find RSA identity key".into()))?;
+            .ok_or_else(|| Error::HandshakeProto("Couldn't find RSA identity key".into()))?;
 
         // Now verify the RSA identity -> Ed Identity crosscert.
         //
@@ -335,15 +429,14 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
         // weak.
         let rsa_cert = c
             .cert_body(CertType::RSA_ID_V_IDENTITY)
-            .ok_or_else(|| Error::ChanProto("No RSA->Ed crosscert".into()))?;
+            .ok_or_else(|| Error::HandshakeProto("No RSA->Ed crosscert".into()))?;
         let rsa_cert = tor_cert::rsa::RsaCrosscert::decode(rsa_cert)?
             .check_signature(&pkrsa)
-            .map_err(|_| Error::ChanProto("Bad RSA->Ed crosscert signature".into()))?
-            .check_valid_at_opt(now)
-            .map_err(|_| Error::ChanProto("RSA->Ed crosscert expired or invalid".into()))?;
+            .map_err(|_| Error::HandshakeProto("Bad RSA->Ed crosscert signature".into()))?;
+        let (rsa_cert_timeliness, rsa_cert) = check_timeliness(rsa_cert, now, self.clock_skew);
 
         if !rsa_cert.subject_key_matches(identity_key) {
-            return Err(Error::ChanProto(
+            return Err(Error::HandshakeProto(
                 "RSA->Ed crosscert certifies incorrect key".into(),
             ));
         }
@@ -366,13 +459,21 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> UnverifiedChannel<T> {
         // usually a different situation than "this peer couldn't even
         // identify itself right."
         if *peer.ed_identity() != ed25519_id {
-            return Err(Error::ChanProto("Peer ed25519 id not as expected".into()));
+            return Err(Error::HandshakeProto(
+                "Peer ed25519 id not as expected".into(),
+            ));
         }
 
         if *peer.rsa_identity() != rsa_id {
-            return Err(Error::ChanProto("Peer RSA id not as expected".into()));
+            return Err(Error::HandshakeProto("Peer RSA id not as expected".into()));
         }
 
+        // We note expired certs last, since any other reason we might reject a
+        // handshake is probably more serious.
+        id_sk_timeliness?;
+        sk_tls_timeliness?;
+        rsa_cert_timeliness?;
+
         Ok(VerifiedChannel {
             link_protocol: self.link_protocol,
             tls: self.tls,
@@ -401,7 +502,10 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> VerifiedChannel<T> {
         crate::note_incoming_traffic();
         trace!("{}: Sending netinfo cell.", self.unique_id);
         let netinfo = msg::Netinfo::for_client(self.target_addr.as_ref().map(SocketAddr::ip));
-        self.tls.send(netinfo.into()).await?;
+        self.tls
+            .send(netinfo.into())
+            .await
+            .map_err(codec_err_to_handshake)?;
 
         debug!(
             "{}: Completed handshake with {} [{}]",
@@ -410,13 +514,18 @@ impl<T: AsyncRead + AsyncWrite + Send + Unpin + 'static> VerifiedChannel<T> {
 
         let (tls_sink, tls_stream) = self.tls.split();
 
+        let peer_id = OwnedChanTarget::new(
+            self.target_addr.into_iter().collect(),
+            self.ed25519_id,
+            self.rsa_id,
+        );
+
         Ok(super::Channel::new(
             self.link_protocol,
             Box::new(tls_sink),
             Box::new(tls_stream),
             self.unique_id,
-            self.ed25519_id,
-            self.rsa_id,
+            peer_id,
         ))
     }
 }
@@ -436,7 +545,13 @@ pub(super) mod test {
     // no certificates in this cell, but connect() doesn't care.
     const NOCERTS: &[u8] = &hex!("00000000 81 0001 00");
     const NETINFO_PREFIX: &[u8] = &hex!(
-        "00000000 08 085F9067F7
+        "00000000 08 00000000
+         04 04 7f 00 00 02
+         01
+         04 04 7f 00 00 03"
+    );
+    const NETINFO_PREFIX_WITH_TIME: &[u8] = &hex!(
+        "00000000 08 48949290
          04 04 7f 00 00 02
          01
          04 04 7f 00 00 03"
@@ -462,18 +577,21 @@ pub(super) mod test {
     #[test]
     fn connect_ok() -> Result<()> {
         tor_rtcompat::test_with_one_runtime!(|_rt| async move {
+            let now = SystemTime::UNIX_EPOCH + Duration::from_secs(1217696400);
             let mut buf = Vec::new();
             // versions cell
             buf.extend_from_slice(VERSIONS);
             // certs cell -- no certs in it, but this function doesn't care.
             buf.extend_from_slice(NOCERTS);
             // netinfo cell -- quite minimal.
-            add_netinfo(&mut buf);
+            add_padded(&mut buf, NETINFO_PREFIX);
             let mb = MsgBuf::new(&buf[..]);
             let handshake = OutboundClientHandshake::new(mb, None);
-            let unverified = handshake.connect().await?;
+            let unverified = handshake.connect(|| now).await?;
 
             assert_eq!(unverified.link_protocol, 4);
+            // No timestamp in the NETINFO, so no skew.
+            assert_eq!(unverified.clock_skew(), ClockSkew::None);
 
             // Try again with an authchallenge cell and some padding.
             let mut buf = Vec::new();
@@ -482,10 +600,22 @@ pub(super) mod test {
             buf.extend_from_slice(VPADDING);
             buf.extend_from_slice(AUTHCHALLENGE);
             buf.extend_from_slice(VPADDING);
-            add_netinfo(&mut buf);
+            add_padded(&mut buf, NETINFO_PREFIX_WITH_TIME);
             let mb = MsgBuf::new(&buf[..]);
             let handshake = OutboundClientHandshake::new(mb, None);
-            let _unverified = handshake.connect().await?;
+            let unverified = handshake.connect(|| now).await?;
+            // Correct timestamp in the NETINFO, so no skew.
+            assert_eq!(unverified.clock_skew(), ClockSkew::None);
+
+            // Now pretend our clock is fast.
+            let now2 = now + Duration::from_secs(3600);
+            let mb = MsgBuf::new(&buf[..]);
+            let handshake = OutboundClientHandshake::new(mb, None);
+            let unverified = handshake.connect(|| now2).await?;
+            assert_eq!(
+                unverified.clock_skew(),
+                ClockSkew::Fast(Duration::from_secs(3600))
+            );
 
             Ok(())
         })
@@ -494,24 +624,24 @@ pub(super) mod test {
     async fn connect_err<T: Into<Vec<u8>>>(input: T) -> Error {
         let mb = MsgBuf::new(input);
         let handshake = OutboundClientHandshake::new(mb, None);
-        handshake.connect().await.err().unwrap()
+        handshake.connect(SystemTime::now).await.err().unwrap()
     }
 
     #[test]
     fn connect_badver() {
         tor_rtcompat::test_with_one_runtime!(|_rt| async move {
             let err = connect_err(&b"HTTP://"[..]).await;
-            assert!(matches!(err, Error::ChanProto(_)));
+            assert!(matches!(err, Error::HandshakeProto(_)));
             assert_eq!(
                 format!("{}", err),
-                "channel protocol violation: Doesn't seem to be a tor relay"
+                "handshake protocol violation: Doesn't seem to be a tor relay"
             );
 
             let err = connect_err(&hex!("0000 07 0004 1234 ffff")[..]).await;
-            assert!(matches!(err, Error::ChanProto(_)));
+            assert!(matches!(err, Error::HandshakeProto(_)));
             assert_eq!(
                 format!("{}", err),
-                "channel protocol violation: No shared link protocols"
+                "handshake protocol violation: No shared link protocols"
             );
         });
     }
@@ -524,10 +654,7 @@ pub(super) mod test {
             // Here's a certs cell that will fail.
             buf.extend_from_slice(&hex!("00000000 81 0001 01")[..]);
             let err = connect_err(buf).await;
-            assert!(matches!(
-                err,
-                Error::CellErr(tor_cell::Error::BytesErr(tor_bytes::Error::Truncated))
-            ));
+            assert!(matches!(err, Error::HandshakeProto(_)));
         });
     }
 
@@ -540,10 +667,10 @@ pub(super) mod test {
             buf.extend_from_slice(NOCERTS);
             add_netinfo(&mut buf);
             let err = connect_err(buf).await;
-            assert!(matches!(err, Error::ChanProto(_)));
+            assert!(matches!(err, Error::HandshakeProto(_)));
             assert_eq!(
                 format!("{}", err),
-                "channel protocol violation: Duplicate certs cell"
+                "handshake protocol violation: Duplicate certs cell"
             );
 
             let mut buf = Vec::new();
@@ -553,10 +680,10 @@ pub(super) mod test {
             buf.extend_from_slice(AUTHCHALLENGE);
             add_netinfo(&mut buf);
             let err = connect_err(buf).await;
-            assert!(matches!(err, Error::ChanProto(_)));
+            assert!(matches!(err, Error::HandshakeProto(_)));
             assert_eq!(
                 format!("{}", err),
-                "channel protocol violation: Duplicate authchallenge cell"
+                "handshake protocol violation: Duplicate authchallenge cell"
             );
         });
     }
@@ -568,10 +695,10 @@ pub(super) mod test {
             buf.extend_from_slice(VERSIONS);
             add_netinfo(&mut buf);
             let err = connect_err(buf).await;
-            assert!(matches!(err, Error::ChanProto(_)));
+            assert!(matches!(err, Error::HandshakeProto(_)));
             assert_eq!(
                 format!("{}", err),
-                "channel protocol violation: Missing certs cell"
+                "handshake protocol violation: Missing certs cell"
             );
         });
     }
@@ -584,10 +711,10 @@ pub(super) mod test {
             // here's a create cell.
             add_padded(&mut buf, &hex!("00000001 01")[..]);
             let err = connect_err(buf).await;
-            assert!(matches!(err, Error::ChanProto(_)));
+            assert!(matches!(err, Error::HandshakeProto(_)));
             assert_eq!(
                 format!("{}", err),
-                "channel protocol violation: Unexpected cell type CREATE"
+                "handshake protocol violation: Unexpected cell type CREATE"
             );
         });
     }
@@ -595,11 +722,13 @@ pub(super) mod test {
     fn make_unverified(certs: msg::Certs) -> UnverifiedChannel<MsgBuf> {
         let localhost = std::net::IpAddr::V4(std::net::Ipv4Addr::LOCALHOST);
         let netinfo_cell = msg::Netinfo::for_client(Some(localhost));
+        let clock_skew = ClockSkew::None;
         UnverifiedChannel {
             link_protocol: 4,
             tls: futures_codec::Framed::new(MsgBuf::new(&b""[..]), ChannelCodec::new(4)),
             certs_cell: certs,
             netinfo_cell,
+            clock_skew,
             target_addr: None,
             unique_id: UniqId::new(),
         }
@@ -654,7 +783,7 @@ pub(super) mod test {
         .unwrap();
         assert_eq!(
             format!("{}", err),
-            "channel protocol violation: Missing IDENTITY_V_SIGNING certificate"
+            "handshake protocol violation: Missing IDENTITY_V_SIGNING certificate"
         );
     }
 
@@ -709,7 +838,7 @@ pub(super) mod test {
 
             assert_eq!(
                 format!("{}", res),
-                format!("channel protocol violation: {}", expect_err.unwrap())
+                format!("handshake protocol violation: {}", expect_err.unwrap())
             );
         }
     }
@@ -733,7 +862,7 @@ pub(super) mod test {
 
         assert_eq!(
             format!("{}", err),
-            "channel protocol violation: Peer ed25519 id not as expected"
+            "handshake protocol violation: Peer ed25519 id not as expected"
         );
 
         let err = certs_test(
@@ -748,7 +877,7 @@ pub(super) mod test {
 
         assert_eq!(
             format!("{}", err),
-            "channel protocol violation: Peer RSA id not as expected"
+            "handshake protocol violation: Peer RSA id not as expected"
         );
 
         let err = certs_test(
@@ -763,7 +892,7 @@ pub(super) mod test {
 
         assert_eq!(
             format!("{}", err),
-            "channel protocol violation: Peer cert did not authenticate TLS cert"
+            "handshake protocol violation: Peer cert did not authenticate TLS cert"
         );
     }
 
@@ -791,7 +920,7 @@ pub(super) mod test {
 
         assert_eq!(
             format!("{}", res),
-            "channel protocol violation: Invalid ed25519 signature in handshake"
+            "handshake protocol violation: Invalid ed25519 signature in handshake"
         );
 
         let mut certs = msg::Certs::new_empty();
@@ -811,7 +940,7 @@ pub(super) mod test {
 
         assert_eq!(
             format!("{}", res),
-            "channel protocol violation: Bad RSA->Ed crosscert signature"
+            "handshake protocol violation: Bad RSA->Ed crosscert signature"
         );
     }
 
