@@ -18,6 +18,7 @@ use futures::channel::{mpsc, oneshot};
 use futures::sink::SinkExt;
 use futures::stream::Stream;
 use futures::Sink;
+use tor_error::internal;
 
 use std::convert::TryInto;
 use std::fmt;
@@ -26,19 +27,28 @@ use std::sync::atomic::Ordering;
 use std::sync::Arc;
 use std::task::Poll;
 
-use crate::channel::{unique_id, ChannelDetails};
+use crate::channel::{codec::CodecError, unique_id, ChannelDetails};
 use crate::circuit::celltypes::{ClientCircChanMsg, CreateResponse};
 use tracing::{debug, trace};
 
 /// A boxed trait object that can provide `ChanCell`s.
 pub(super) type BoxedChannelStream =
-    Box<dyn Stream<Item = std::result::Result<ChanCell, tor_cell::Error>> + Send + Unpin + 'static>;
+    Box<dyn Stream<Item = std::result::Result<ChanCell, CodecError>> + Send + Unpin + 'static>;
 /// A boxed trait object that can sink `ChanCell`s.
 pub(super) type BoxedChannelSink =
-    Box<dyn Sink<ChanCell, Error = tor_cell::Error> + Send + Unpin + 'static>;
+    Box<dyn Sink<ChanCell, Error = CodecError> + Send + Unpin + 'static>;
 /// The type of a oneshot channel used to inform reactor users of the result of an operation.
 pub(super) type ReactorResultChannel<T> = oneshot::Sender<Result<T>>;
 
+/// Convert `err` to an Error, under the assumption that it's happening on an
+/// open channel.
+fn codec_err_to_chan(err: CodecError) -> Error {
+    match err {
+        CodecError::Io(e) => crate::Error::ChanIoErr(Arc::new(e)),
+        CodecError::Cell(e) => e.into(),
+    }
+}
+
 /// A message telling the channel reactor to do something.
 #[derive(Debug)]
 pub(super) enum CtrlMsg {
@@ -145,7 +155,7 @@ impl Reactor {
 
             // See if the output sink can have cells written to it yet.
             if let Poll::Ready(ret) = Pin::new(&mut self.output).poll_ready(cx) {
-                let _ = ret.map_err(Error::CellErr)?;
+                let _ = ret.map_err(codec_err_to_chan)?;
                 // If it can, check whether we have any cells to send it from `Channel` senders.
                 if let Poll::Ready(msg) = Pin::new(&mut self.cells).poll_next(cx) {
                     match msg {
@@ -172,7 +182,7 @@ impl Reactor {
             if let Poll::Ready(ret) = Pin::new(&mut self.input).poll_next(cx) {
                 match ret {
                     None => return Poll::Ready(Err(ReactorError::Shutdown)),
-                    Some(r) => input = Some(r.map_err(Error::CellErr)?),
+                    Some(r) => input = Some(r.map_err(codec_err_to_chan)?),
                 }
             }
 
@@ -180,7 +190,7 @@ impl Reactor {
             // we just want to keep flushing it (hence the _).
             let _ = Pin::new(&mut self.output)
                 .poll_flush(cx)
-                .map_err(Error::CellErr)?;
+                .map_err(codec_err_to_chan)?;
 
             // If all three values aren't present, return Pending and wait to get polled again
             // so that one of them is present.
@@ -201,11 +211,11 @@ impl Reactor {
         if let Some(cts) = cell_to_send {
             Pin::new(&mut self.output)
                 .start_send(cts)
-                .map_err(Error::CellErr)?;
+                .map_err(codec_err_to_chan)?;
             // Give the sink a little flush, to make sure it actually starts doing things.
             futures::future::poll_fn(|cx| Pin::new(&mut self.output).poll_flush(cx))
                 .await
-                .map_err(Error::CellErr)?;
+                .map_err(codec_err_to_chan)?;
         }
         Ok(()) // Run again.
     }
@@ -229,6 +239,7 @@ impl Reactor {
                     .add_ent(&mut rng, created_sender, sender)
                     .map(|id| (id, circ_unique_id));
                 let _ = tx.send(ret); // don't care about other side going away
+                self.update_disused_since();
             }
         }
         Ok(())
@@ -286,20 +297,25 @@ impl Reactor {
 
     /// Give the RELAY cell `msg` to the appropriate circuit.
     async fn deliver_relay(&mut self, circid: CircId, msg: ChanMsg) -> Result<()> {
-        match self.circs.get_mut(circid) {
-            Some(CircEnt::Open(s)) => {
+        let mut ent = self
+            .circs
+            .get_mut(circid)
+            .ok_or_else(|| Error::ChanProto("Relay cell on nonexistent circuit".into()))?;
+
+        match &mut *ent {
+            CircEnt::Open(s) => {
                 // There's an open circuit; we can give it the RELAY cell.
                 if s.send(msg.try_into()?).await.is_err() {
+                    drop(ent);
                     // The circuit's receiver went away, so we should destroy the circuit.
                     self.outbound_destroy_circ(circid).await?;
                 }
                 Ok(())
             }
-            Some(CircEnt::Opening(_, _)) => Err(Error::ChanProto(
+            CircEnt::Opening(_, _) => Err(Error::ChanProto(
                 "Relay cell on pending circuit before CREATED* received".into(),
             )),
-            Some(CircEnt::DestroySent(hs)) => hs.receive_cell(),
-            None => Err(Error::ChanProto("Relay cell on nonexistent circuit".into())),
+            CircEnt::DestroySent(hs) => hs.receive_cell(),
         }
     }
 
@@ -311,7 +327,9 @@ impl Reactor {
         // TODO(nickm) I think that this one actually means the other side
         // is closed. See arti#269.
         target.send(created).map_err(|_| {
-            Error::InternalError("Circuit queue rejected created message. Is it closing?".into())
+            Error::from(internal!(
+                "Circuit queue rejected created message. Is it closing?"
+            ))
         })
     }
 
@@ -320,6 +338,7 @@ impl Reactor {
     async fn deliver_destroy(&mut self, circid: CircId, msg: ChanMsg) -> Result<()> {
         // Remove the circuit from the map: nothing more can be done with it.
         let entry = self.circs.remove(circid);
+        self.update_disused_since();
         match entry {
             // If the circuit is waiting for CREATED, tell it that it
             // won't get one.
@@ -330,9 +349,7 @@ impl Reactor {
                     // TODO(nickm) I think that this one actually means the other side
                     // is closed. See arti#269.
                     .map_err(|_| {
-                        Error::InternalError(
-                            "pending circuit wasn't interested in Destroy cell?".into(),
-                        )
+                        internal!("pending circuit wasn't interested in destroy cell?").into()
                     })
             }
             // It's an open circuit: tell it that it got a DESTROY cell.
@@ -343,7 +360,7 @@ impl Reactor {
                     // TODO(nickm) I think that this one actually means the other side
                     // is closed. See arti#269.
                     .map_err(|_| {
-                        Error::InternalError("circuit wasn't interested in destroy cell?".into())
+                        internal!("open circuit wasn't interested in destroy cell?").into()
                     })
             }
             // We've sent a destroy; we can leave this circuit removed.
@@ -358,7 +375,7 @@ impl Reactor {
 
     /// Helper: send a cell on the outbound sink.
     async fn send_cell(&mut self, cell: ChanCell) -> Result<()> {
-        self.output.send(cell).await?;
+        self.output.send(cell).await.map_err(codec_err_to_chan)?;
         Ok(())
     }
 
@@ -371,12 +388,24 @@ impl Reactor {
         // TODO: It would be great to have a tighter upper bound for
         // the number of relay cells we'll receive.
         self.circs.destroy_sent(id, HalfCirc::new(3000));
+        self.update_disused_since();
         let destroy = Destroy::new(DestroyReason::NONE).into();
         let cell = ChanCell::new(id, destroy);
         self.send_cell(cell).await?;
 
         Ok(())
     }
+
+    /// Update disused timestamp with current time if this channel is no longer used
+    fn update_disused_since(&self) {
+        if self.circs.open_ent_count() == 0 {
+            // Update disused_since if it still indicates that the channel is in use
+            self.details.unused_since.update_if_none();
+        } else {
+            // Mark this channel as in use
+            self.details.unused_since.clear();
+        }
+    }
 }
 
 #[cfg(test)]
@@ -388,8 +417,9 @@ pub(crate) mod test {
     use futures::sink::SinkExt;
     use futures::stream::StreamExt;
     use futures::task::SpawnExt;
+    use tor_linkspec::OwnedChanTarget;
 
-    type CodecResult = std::result::Result<ChanCell, tor_cell::Error>;
+    type CodecResult = std::result::Result<ChanCell, CodecError>;
 
     pub(crate) fn new_reactor() -> (
         crate::channel::Channel,
@@ -401,19 +431,17 @@ pub(crate) mod test {
         let (send1, recv1) = mpsc::channel(32);
         let (send2, recv2) = mpsc::channel(32);
         let unique_id = UniqId::new();
-        let ed_id = [6; 32].into();
-        let rsa_id = [10; 20].into();
+        let dummy_target = OwnedChanTarget::new(vec![], [6; 32].into(), [10; 20].into());
         let send1 = send1.sink_map_err(|e| {
             trace!("got sink error: {}", e);
-            tor_cell::Error::ChanProto("dummy message".into())
+            CodecError::Cell(tor_cell::Error::ChanProto("dummy message".into()))
         });
         let (chan, reactor) = crate::channel::Channel::new(
             link_protocol,
             Box::new(send1),
             Box::new(recv2),
             unique_id,
-            ed_id,
-            rsa_id,
+            dummy_target,
         );
         (chan, reactor, recv1, send2)
     }
@@ -462,6 +490,7 @@ pub(crate) mod test {
     fn new_circ_closed() {
         tor_rtcompat::test_with_all_runtimes!(|rt| async move {
             let (chan, mut reactor, mut output, _input) = new_reactor();
+            assert!(chan.duration_unused().is_some()); // unused yet
 
             let (ret, reac) = futures::join!(chan.new_circ(), reactor.run_once());
             let (pending, circr) = ret.unwrap();
@@ -474,17 +503,20 @@ pub(crate) mod test {
             let id = pending.peek_circid();
 
             let ent = reactor.circs.get_mut(id);
-            assert!(matches!(ent, Some(CircEnt::Opening(_, _))));
+            assert!(matches!(*ent.unwrap(), CircEnt::Opening(_, _)));
+            assert!(chan.duration_unused().is_none()); // in use
+
             // Now drop the circuit; this should tell the reactor to remove
             // the circuit from the map.
             drop(pending);
 
             reactor.run_once().await.unwrap();
             let ent = reactor.circs.get_mut(id);
-            assert!(matches!(ent, Some(CircEnt::DestroySent(_))));
+            assert!(matches!(*ent.unwrap(), CircEnt::DestroySent(_)));
             let cell = output.next().await.unwrap();
             assert_eq!(cell.circid(), id);
             assert!(matches!(cell.msg(), ChanMsg::Destroy(_)));
+            assert!(chan.duration_unused().is_some()); // unused again
         });
     }
 
@@ -512,7 +544,7 @@ pub(crate) mod test {
             let id = pending.peek_circid();
 
             let ent = reactor.circs.get_mut(id);
-            assert!(matches!(ent, Some(CircEnt::Opening(_, _))));
+            assert!(matches!(*ent.unwrap(), CircEnt::Opening(_, _)));
 
             #[allow(clippy::clone_on_copy)]
             let rtc = rt.clone();
@@ -528,18 +560,17 @@ pub(crate) mod test {
             let (circ, _) =
                 futures::join!(pending.create_firsthop_fast(&circparams), send_response);
             // Make sure statuses are as expected.
-            assert!(matches!(circ.err().unwrap(), Error::BadHandshake));
+            assert!(matches!(circ.err().unwrap(), Error::BadCircHandshake));
 
             reactor.run_once().await.unwrap();
 
             // Make sure that the createfast cell got sent
             let cell_sent = output.next().await.unwrap();
-            dbg!(cell_sent.msg());
             assert!(matches!(cell_sent.msg(), msg::ChanMsg::CreateFast(_)));
 
             // But the next run if the reactor will make the circuit get closed.
             let ent = reactor.circs.get_mut(id);
-            assert!(matches!(ent, Some(CircEnt::DestroySent(_))));
+            assert!(matches!(*ent.unwrap(), CircEnt::DestroySent(_)));
         });
     }
 
