@@ -22,12 +22,13 @@ use tor_cell::relaycell::{RelayCell, RelayCmd, StreamId};
 use futures::channel::{mpsc, oneshot};
 use futures::Sink;
 use futures::Stream;
+use tor_error::internal;
 
-use std::sync::atomic::{AtomicU8, Ordering};
 use std::sync::Arc;
 use std::task::{Context, Poll};
 
 use crate::channel::Channel;
+use crate::circuit::path;
 #[cfg(test)]
 use crate::circuit::sendme::CircTag;
 use crate::circuit::sendme::StreamSendWindow;
@@ -35,7 +36,7 @@ use crate::crypto::handshake::ntor::{NtorClient, NtorPublicKey};
 use crate::crypto::handshake::{ClientHandshake, KeyGenerator};
 use tor_cell::chancell;
 use tor_cell::chancell::{ChanCell, CircId};
-use tor_linkspec::LinkSpec;
+use tor_linkspec::{LinkSpec, OwnedChanTarget};
 use tor_llcrypto::pk;
 use tracing::{debug, trace, warn};
 
@@ -87,6 +88,10 @@ pub(super) enum CtrlMsg {
     },
     /// Extend a circuit by one hop, using the ntor handshake.
     ExtendNtor {
+        /// The peer that we're extending to.
+        ///
+        /// Used to extend our record of the circuit's path.
+        peer_id: OwnedChanTarget,
         /// The handshake type to use for this hop.
         public_key: NtorPublicKey,
         /// Information about how to connect to the relay we're extending to.
@@ -168,12 +173,16 @@ pub(super) struct CircHop {
     sendwindow: sendme::CircSendWindow,
     /// Buffer for messages we can't send to this hop yet due to congestion control.
     ///
-    /// Contains the tag we should give to the send window, and the cell to send.
+    /// Contains the cell to send, and a boolean equivalent to the `early` parameter
+    /// in `Reactor::send_relay_cell` (as in, whether to send the cell using `RELAY_EARLY`).
     ///
     /// This shouldn't grow unboundedly: we try and pop things off it first before
     /// doing things that would result in it growing (and stop before growing it
     /// if popping things off it can't be done).
-    outbound: VecDeque<([u8; 20], ChanCell)>,
+    ///
+    /// NOTE: Control messages could potentially add unboundedly to this, although that's
+    ///       not likely to happen (and isn't triggereable from the network, either).
+    outbound: VecDeque<(bool, RelayCell)>,
 }
 
 /// Enumeration to determine whether we require circuit-level SENDME cells to be
@@ -262,6 +271,10 @@ struct CircuitExtender<H, L, FWD, REV>
 where
     H: ClientHandshake,
 {
+    /// The peer that we're extending to.
+    ///
+    /// Used to extend our record of the circuit's path.
+    peer_id: OwnedChanTarget,
     /// Handshake state.
     state: Option<H::StateType>,
     /// Whether the hop supports authenticated SENDME cells.
@@ -297,8 +310,10 @@ where
     /// goes along with the handshake, and the `linkspecs` are the
     /// link specifiers to include in the EXTEND cell to tell the
     /// current last hop which relay to connect to.
+    #[allow(clippy::too_many_arguments)]
     fn begin(
         cx: &mut Context<'_>,
+        peer_id: OwnedChanTarget,
         handshake_id: u16,
         key: &H::KeyType,
         linkspecs: Vec<LinkSpec>,
@@ -335,6 +350,7 @@ where
         // ... and now we wait for a response.
 
         Ok(Self {
+            peer_id,
             state: Some(state),
             require_sendme_auth,
             params,
@@ -371,7 +387,13 @@ where
 
         let msg = match msg {
             RelayMsg::Extended2(e) => e,
-            _ => return Err(Error::InternalError("Body didn't match cmd".into())),
+            _ => {
+                return Err(Error::from(internal!(
+                    "Message body {:?} didn't match cmd {:?}",
+                    msg,
+                    msg.cmd()
+                )))
+            }
         };
         let relay_handshake = msg.into_body();
 
@@ -394,6 +416,7 @@ where
         // If we get here, it succeeded.  Add a new hop to the circuit.
         let (layer_fwd, layer_back) = layer.split();
         reactor.add_hop(
+            self.peer_id.clone(),
             self.require_sendme_auth,
             Box::new(layer_fwd),
             Box::new(layer_back),
@@ -413,8 +436,12 @@ pub struct Reactor {
     pub(super) control: mpsc::UnboundedReceiver<CtrlMsg>,
     /// Buffer for cells we can't send out the channel yet due to it being full.
     ///
-    /// This should be used very very rarely: see `send_msg_direct`'s comments for more
-    /// information. (in fact, using it will generate a warning!)
+    /// We try and dequeue off this first before doing anything else, ensuring that
+    /// it cannot grow unboundedly (and if we start having to enqueue things on here after
+    /// the channel shows backpressure, we stop pulling from receivers that could send here).
+    ///
+    /// NOTE: Control messages could potentially add unboundedly to this, although that's
+    ///       not likely to happen (and isn't triggereable from the network, either).
     pub(super) outbound: VecDeque<ChanCell>,
     /// The channel this circuit is using to send cells through.
     pub(super) channel: Channel,
@@ -431,7 +458,7 @@ pub struct Reactor {
     /// List of hops state objects used by the reactor
     pub(super) hops: Vec<CircHop>,
     /// Shared atomic for the number of hops this circuit has.
-    pub(super) num_hops: Arc<AtomicU8>,
+    pub(super) path: Arc<path::Path>,
     /// An identifier for logging about this reactor's circuit.
     pub(super) unique_id: UniqId,
     /// This circuit's identifier on the upstream channel.
@@ -538,28 +565,29 @@ impl Reactor {
                     }
 
                     // Let's look at our hops, and streams for each hop.
-                    for (i, hop) in self.hops.iter_mut().enumerate() {
+                    for i in 0..self.hops.len() {
                         let hop_num = HopNum::from(i as u8);
                         // If we can, drain our queue of things we tried to send earlier, but
                         // couldn't due to congestion control.
-                        if hop.sendwindow.window() > 0 {
-                            'hop: while let Some((tag, msg)) = hop.outbound.pop_front() {
+                        if self.hops[i].sendwindow.window() > 0 {
+                            'hop: while let Some((early, cell)) = self.hops[i].outbound.pop_front()
+                            {
                                 trace!(
                                     "{}: sending from hop-{}-enqueued: {:?}",
                                     self.unique_id,
                                     i,
-                                    msg
+                                    cell
                                 );
-                                Pin::new(&mut self.channel).start_send(msg)?;
-                                hop.sendwindow.take(&tag)?;
+                                self.send_relay_cell(cx, hop_num, early, cell)?;
                                 if !self.channel.poll_ready(cx)? {
                                     break 'outer;
                                 }
-                                if hop.sendwindow.window() == 0 {
+                                if self.hops[i].sendwindow.window() == 0 {
                                     break 'hop;
                                 }
                             }
                         }
+                        let hop = &mut self.hops[i];
                         // Look at all of the streams on this hop.
                         for (id, stream) in hop.map.inner().iter_mut() {
                             if let StreamEnt::Open {
@@ -706,7 +734,10 @@ impl Reactor {
         debug!("{}: Handshake complete; circuit created.", self.unique_id);
 
         let (layer_fwd, layer_back) = layer.split();
+        let peer_id = self.channel.target().clone();
+
         self.add_hop(
+            peer_id,
             require_sendme_auth,
             Box::new(layer_fwd),
             Box::new(layer_back),
@@ -783,6 +814,7 @@ impl Reactor {
     /// Add a hop to the end of this circuit.
     fn add_hop(
         &mut self,
+        peer_id: OwnedChanTarget,
         require_sendme_auth: RequireSendmeAuth,
         fwd: Box<dyn OutboundClientLayer + 'static + Send>,
         rev: Box<dyn InboundClientLayer + 'static + Send>,
@@ -795,7 +827,7 @@ impl Reactor {
         self.hops.push(hop);
         self.crypto_in.add_layer(rev);
         self.crypto_out.add_layer(fwd);
-        self.num_hops.fetch_add(1, Ordering::SeqCst);
+        self.path.push_hop(peer_id);
     }
 
     /// Handle a RELAY cell on this circuit with stream ID 0.
@@ -897,23 +929,31 @@ impl Reactor {
     /// that would send here while you know you're unable to forward the messages on).
     fn send_msg_direct(&mut self, cx: &mut Context<'_>, msg: ChanMsg) -> Result<()> {
         let cell = ChanCell::new(self.channel_id, msg);
-        if self.channel.poll_ready(cx)? {
+        // NOTE(eta): We need to check whether the outbound queue is empty before trying to send:
+        //            if we just checked whether the channel was ready, it'd be possible for
+        //            cells to be sent out of order, since it could transition from not ready to
+        //            ready during one cycle of the reactor!
+        //            (This manifests as a protocol violation.)
+        if self.outbound.is_empty() && self.channel.poll_ready(cx)? {
             Pin::new(&mut self.channel).start_send(cell)?;
         } else {
-            // This case shouldn't actually happen that often, if ever. We generally check whether
-            // the channel can be sent to before calling this function (the one exception at the
-            // time of writing is in circuit creation).
+            // This has been observed to happen in code that doesn't have bugs in it, simply due
+            // to the way `Channel`'s `poll_ready` implementation works (it can change due to
+            // the actions of another thread in between callers of this function checking it,
+            // and this function checking it).
             //
-            // If this is suddenly getting hit and it wasn't before, maybe you added something that
-            // doesn't bother to check the channel (`self.channel.poll_ready(cx)`) before calling
-            // this function, and that's getting used a lot?
-            //
-            // We don't want to drop cells on the floor, though, so this is good to have.
-            warn!(
+            // However, if it's happening a lot more than it used to, that probably indicates
+            // some caller that's not checking whether the channel is full before calling
+            // this function.
+
+            debug!(
                 "{}: having to enqueue cell due to backpressure: {:?}",
                 self.unique_id, cell
             );
             self.outbound.push_back(cell);
+
+            // Ensure we absolutely get scheduled again to clear `self.outbound`.
+            cx.waker().wake_by_ref();
         }
         Ok(())
     }
@@ -947,8 +987,30 @@ impl Reactor {
     ) -> Result<()> {
         let c_t_w = sendme::cell_counts_towards_windows(&cell);
         let stream_id = cell.stream_id();
+        // Check whether the hop send window is empty, if this cell counts towards windows.
+        // NOTE(eta): It is imperative this happens *before* calling encrypt() below, otherwise
+        //            we'll have cells rejected due to a protocol violation! (Cells have to be
+        //            sent out in the order they were passed to encrypt().)
+        if c_t_w {
+            let hop_num = Into::<usize>::into(hop);
+            let hop = &mut self.hops[hop_num];
+            if hop.sendwindow.window() == 0 {
+                // Send window is empty! Push this cell onto the hop's outbound queue, and it'll
+                // get sent later.
+                trace!(
+                    "{}: having to use onto hop {} queue for cell: {:?}",
+                    self.unique_id,
+                    hop_num,
+                    cell
+                );
+                hop.outbound.push_back((early, cell));
+                return Ok(());
+            }
+        }
         let mut body: RelayCellBody = cell.encode(&mut rand::thread_rng())?.into();
         let tag = self.crypto_out.encrypt(&mut body, hop)?;
+        // NOTE(eta): Now that we've encrypted the cell, we *must* either send it or abort
+        //            the whole circuit (e.g. by returning an error).
         let msg = chancell::msg::Relay::from_raw(body.into());
         let msg = if early {
             ChanMsg::RelayEarly(msg)
@@ -960,19 +1022,7 @@ impl Reactor {
         if c_t_w {
             let hop_num = Into::<usize>::into(hop);
             let hop = &mut self.hops[hop_num];
-            if hop.sendwindow.window() == 0 {
-                let cell = ChanCell::new(self.channel_id, msg);
-                // Send window is empty! Push this cell onto the hop's outbound queue, and it'll
-                // get sent later.
-                trace!(
-                    "{}: having to use onto hop {} queue for cell: {:?}",
-                    self.unique_id,
-                    hop_num,
-                    cell
-                );
-                hop.outbound.push_back((*tag, cell));
-                return Ok(());
-            }
+            // checked by earlier conditional, so this shouldn't fail
             hop.sendwindow.take(tag)?;
             if !stream_id.is_zero() {
                 // We need to decrement the stream-level sendme window.
@@ -1006,9 +1056,9 @@ impl Reactor {
             self.meta_handler = Some((handler, done));
             Ok(())
         } else {
-            Err(Error::InternalError(
-                "Tried to install a meta-cell handler before the old one was gone.".into(),
-            ))
+            Err(Error::from(internal!(
+                "Tried to install a meta-cell handler before the old one was gone."
+            )))
         }
     }
 
@@ -1021,6 +1071,7 @@ impl Reactor {
             // This is handled earlier, since it requires generating a ReactorError.
             CtrlMsg::Shutdown => panic!("got a CtrlMsg::Shutdown in handle_control"),
             CtrlMsg::ExtendNtor {
+                peer_id,
                 public_key,
                 linkspecs,
                 require_sendme_auth,
@@ -1029,6 +1080,7 @@ impl Reactor {
             } => {
                 match CircuitExtender::<NtorClient, Tor1RelayCrypto, _, _>::begin(
                     cx,
+                    peer_id,
                     0x02,
                     &public_key,
                     linkspecs,
@@ -1076,9 +1128,10 @@ impl Reactor {
                     RequireSendmeAuth::No
                 };
 
+                let dummy_peer_id = OwnedChanTarget::new(vec![], [4; 32].into(), [5; 20].into());
                 let fwd = Box::new(DummyCrypto::new(fwd_lasthop));
                 let rev = Box::new(DummyCrypto::new(rev_lasthop));
-                self.add_hop(require_sendme_auth, fwd, rev, &params);
+                self.add_hop(dummy_peer_id, require_sendme_auth, fwd, rev, &params);
                 let _ = done.send(Ok(()));
             }
             #[cfg(test)]
@@ -1086,9 +1139,10 @@ impl Reactor {
                 let _ = done.send(if let Some(hop) = self.hop_mut(hop) {
                     Ok(hop.sendwindow.window_and_expected_tags())
                 } else {
-                    Err(Error::InternalError(
-                        "received QuerySendWindow for unknown hop".into(),
-                    ))
+                    Err(Error::from(internal!(
+                        "received QuerySendWindow for unknown hop {:?}",
+                        hop
+                    )))
                 });
             }
             #[cfg(test)]
@@ -1111,7 +1165,7 @@ impl Reactor {
     ) -> Result<StreamId> {
         let hop = self
             .hop_mut(hopnum)
-            .ok_or_else(|| Error::InternalError(format!("No such hop {:?}", hopnum)))?;
+            .ok_or_else(|| Error::from(internal!("No such hop {:?}", hopnum)))?;
         let send_window = StreamSendWindow::new(SEND_WINDOW_INIT);
         let r = hop.map.add_ent(sender, rx, send_window)?;
         let cell = RelayCell::new(r, message);
@@ -1126,7 +1180,10 @@ impl Reactor {
     fn close_stream(&mut self, cx: &mut Context<'_>, hopnum: HopNum, id: StreamId) -> Result<()> {
         // Mark the stream as closing.
         let hop = self.hop_mut(hopnum).ok_or_else(|| {
-            Error::InternalError("Tried to close a stream on a hop that wasn't there?".into())
+            Error::from(internal!(
+                "Tried to close a stream on a hop {:?} that wasn't there?",
+                hopnum
+            ))
         })?;
 
         let should_send_end = hop.map.terminate(id)?;
@@ -1182,7 +1239,7 @@ impl Reactor {
             let mut tag_copy = [0_u8; 20];
             // TODO(nickm): This could crash if the tag length changes.  We'll
             // have to refactor it then.
-            (&mut tag_copy).copy_from_slice(tag);
+            tag_copy.copy_from_slice(tag);
             tag_copy
         };
         // Decode the cell.
@@ -1211,7 +1268,10 @@ impl Reactor {
             self.send_relay_cell(cx, hopnum, false, cell)?;
             self.hop_mut(hopnum)
                 .ok_or_else(|| {
-                    Error::InternalError("Trying to send SENDME to nonexistent hop".to_string())
+                    Error::from(internal!(
+                        "Trying to send SENDME to nonexistent hop {:?}",
+                        hopnum
+                    ))
                 })?
                 .recvwindow
                 .put();
