@@ -1,7 +1,8 @@
 //! Define an error type for the tor-proto crate.
-use std::sync::Arc;
+use std::{sync::Arc, time::Duration};
 use thiserror::Error;
 use tor_cell::relaycell::msg::EndReason;
+use tor_error::{ErrorKind, HasKind};
 
 /// An error type for the tor-proto crate.
 ///
@@ -15,55 +16,63 @@ pub enum Error {
     /// object.
     #[error("parsing error: {0}")]
     BytesErr(#[from] tor_bytes::Error),
-    /// An error that occurred from the io system.
-    #[error("io error: {0}")]
-    IoErr(#[source] Arc<std::io::Error>),
+    /// An error that occurred from the io system when using a
+    /// channel.
+    #[error("io error on channel: {0}")]
+    ChanIoErr(#[source] Arc<std::io::Error>),
+    /// An error from the io system that occurred when trying to connect a channel.
+    #[error("io error in handshake: {0}")]
+    HandshakeIoErr(#[source] Arc<std::io::Error>),
     /// An error occurred in the cell-handling layer.
     #[error("cell encoding error: {0}")]
     CellErr(#[source] tor_cell::Error),
-    /// Somebody asked for a key that we didn't have.
-    #[error("specified key was missing")]
-    MissingKey,
     /// We tried to produce too much output for some function.
     #[error("couldn't produce that much output")]
     InvalidOutputLength,
     /// We tried to encrypt a message to a hop that wasn't there.
     #[error("tried to encrypt to nonexistent hop")]
     NoSuchHop,
-    /// There was a programming error somewhere in the code.
-    #[error("Internal programming error: {0}")]
-    InternalError(String),
     /// The authentication information on this cell was completely wrong,
     /// or the cell was corrupted.
     #[error("bad relay cell authentication")]
     BadCellAuth,
     /// A circuit-extension handshake failed.
     #[error("handshake failed")]
-    BadHandshake,
-    /// Protocol violation at the channel level
+    BadCircHandshake,
+    /// Handshake protocol violation.
+    #[error("handshake protocol violation: {0}")]
+    HandshakeProto(String),
+    /// Handshake broken, maybe due to clock skew.
+    ///
+    /// (If the problem can't be due to clock skew, we return HandshakeProto
+    /// instead.)
+    #[error("handshake failed due to expired certificates (possible clock skew)")]
+    HandshakeCertsExpired {
+        /// For how long has the circuit been expired?
+        expired_by: Duration,
+        /// How fast does the relay claim that our clock is?
+        skew: Duration,
+    },
+    /// Protocol violation at the channel level, other than at the handshake
+    /// stage.
     #[error("channel protocol violation: {0}")]
     ChanProto(String),
     /// Protocol violation at the circuit level
     #[error("circuit protocol violation: {0}")]
     CircProto(String),
-    /// Circuit destroyed or channel closed.
-    #[error("circuit destroyed: {0}")]
-    CircDestroy(String),
     /// Channel is closed.
     #[error("channel closed")]
     ChannelClosed,
     /// Circuit is closed.
     #[error("circuit closed")]
     CircuitClosed,
-    /// Stream has ended.
-    #[error("stream ended")]
-    StreamEnded,
     /// Can't allocate any more circuit or stream IDs on a channel.
     #[error("too many entries in map: can't allocate ID")]
     IdRangeFull,
-    /// Couldn't extend a circuit.
+    /// Couldn't extend a circuit because the extending relay or the
+    /// target relay refused our request.
     #[error("circuit extension handshake error: {0}")]
-    CircExtend(&'static str),
+    CircRefused(&'static str),
     /// Tried to make or use a stream to an invalid destination address.
     #[error("invalid stream target address")]
     BadStreamAddress,
@@ -76,12 +85,13 @@ pub enum Error {
     /// Stream protocol violation
     #[error("stream protocol violation: {0}")]
     StreamProto(String),
+
     /// Channel does not match target
     #[error("channel mismatch: {0}")]
     ChanMismatch(String),
-    /// Tried to configure an impossible value
-    #[error("bad configuration value: {0}")]
-    BadConfig(String),
+    /// There was a programming error somewhere in our code, or the calling code.
+    #[error("Programming error: {0}")]
+    Bug(#[from] tor_error::Bug),
     /// Remote DNS lookup failed.
     #[error("remote resolve failed: {0}")]
     ResolveError(String),
@@ -96,18 +106,12 @@ impl From<tor_cell::Error> for Error {
     }
 }
 
-impl From<std::io::Error> for Error {
-    fn from(err: std::io::Error) -> Error {
-        Error::IoErr(Arc::new(err))
-    }
-}
-
 impl From<Error> for std::io::Error {
     fn from(err: Error) -> std::io::Error {
         use std::io::ErrorKind;
         use Error::*;
         let kind = match err {
-            IoErr(e) => match Arc::try_unwrap(e) {
+            ChanIoErr(e) | HandshakeIoErr(e) => match Arc::try_unwrap(e) {
                 Ok(e) => return e,
                 Err(arc) => return std::io::Error::new(arc.kind(), arc),
             },
@@ -118,21 +122,60 @@ impl From<Error> for std::io::Error {
 
             EndReceived(end_reason) => end_reason.into(),
 
-            CircDestroy(_) | ChannelClosed | CircuitClosed | StreamEnded => {
-                ErrorKind::ConnectionReset
-            }
+            ChannelClosed | CircuitClosed => ErrorKind::ConnectionReset,
 
-            BytesErr(_) | MissingKey | BadCellAuth | BadHandshake | ChanProto(_) | CircProto(_)
-            | CellErr(_) | ChanMismatch(_) | StreamProto(_) => ErrorKind::InvalidData,
+            BytesErr(_)
+            | BadCellAuth
+            | BadCircHandshake
+            | HandshakeProto(_)
+            | ChanProto(_)
+            | HandshakeCertsExpired { .. }
+            | CircProto(_)
+            | CellErr(_)
+            | ChanMismatch(_)
+            | StreamProto(_) => ErrorKind::InvalidData,
 
-            InternalError(_) | IdRangeFull | CircExtend(_) | BadConfig(_) | ResolveError(_) => {
-                ErrorKind::Other
-            }
+            Bug(ref e) if e.kind() == tor_error::ErrorKind::BadApiUsage => ErrorKind::InvalidData,
+
+            IdRangeFull | CircRefused(_) | ResolveError(_) | Bug(_) => ErrorKind::Other,
         };
         std::io::Error::new(kind, err)
     }
 }
 
+impl HasKind for Error {
+    fn kind(&self) -> ErrorKind {
+        use tor_bytes::Error as BytesError;
+        use Error as E;
+        use ErrorKind as EK;
+        match self {
+            E::BytesErr(BytesError::Bug(e)) => e.kind(),
+            E::BytesErr(_) => EK::TorProtocolViolation,
+            E::ChanIoErr(_) => EK::LocalNetworkError,
+            E::HandshakeIoErr(_) => EK::TorAccessFailed,
+            E::CellErr(e) => e.kind(),
+            E::InvalidOutputLength => EK::Internal,
+            E::NoSuchHop => EK::BadApiUsage,
+            E::BadCellAuth => EK::TorProtocolViolation,
+            E::BadCircHandshake => EK::TorProtocolViolation,
+            E::HandshakeProto(_) => EK::TorAccessFailed,
+            E::HandshakeCertsExpired { .. } => EK::ClockSkew,
+            E::ChanProto(_) => EK::TorProtocolViolation,
+            E::CircProto(_) => EK::TorProtocolViolation,
+            E::ChannelClosed | E::CircuitClosed => EK::CircuitCollapse,
+            E::IdRangeFull => EK::BadApiUsage,
+            E::CircRefused(_) => EK::CircuitRefused,
+            E::BadStreamAddress => EK::BadApiUsage,
+            E::EndReceived(reason) => reason.kind(),
+            E::NotConnected => EK::BadApiUsage,
+            E::StreamProto(_) => EK::TorProtocolViolation,
+            E::ChanMismatch(_) => EK::RelayIdMismatch,
+            E::ResolveError(_) => EK::RemoteHostNotFound,
+            E::Bug(e) => e.kind(),
+        }
+    }
+}
+
 /// Internal type: Error return value from reactor's run_once
 /// function: indicates an error or a shutdown.
 #[derive(Debug)]
